<?php

namespace QUI\ERP\Payments\Stripe\PaymentMethods\Recurring;

use Exception;
use QUI;
use QUI\ERP\Accounting\Invoice\Invoice;
use QUI\ERP\Order\AbstractOrder;
use QUI\ERP\Payments\Stripe\AbstractBasePayment;
use QUI\ERP\Payments\Stripe\StripeException;
use Stripe\Exception\ApiErrorException;
use Stripe\PaymentIntent as StripePaymentIntent;
use QUI\ERP\Accounting\Payments\Types\RecurringPaymentInterface;
use QUI\ERP\Payments\Stripe\Provider;
use Stripe\Subscription as StripeSubscription;
use QUI\ERP\Payments\Stripe\Utils;

use function array_column;

/**
 * Class BaseRecurringPayment
 *
 * Stripe payment for recurring payments
 */
abstract class AbstractBaseRecurringPayment extends AbstractBasePayment implements RecurringPaymentInterface
{
    /**
     * Return attributes for creating a PaymentIntent
     *
     * @param AbstractOrder $Order
     * @param string $paymentMethodId - Stripe PaymentMethod ID
     * @return StripePaymentIntent
     *
     * @throws ApiErrorException
     * @throws QUI\Exception
     */
    protected function createPaymentIntentForOrder(AbstractOrder $Order, string $paymentMethodId): StripePaymentIntent
    {
        return StripePaymentIntent::create([
            'payment_method' => $paymentMethodId,
            'amount' => Utils::getCentAmount($Order),
            'currency' => mb_strtolower($Order->getCurrency()->getCode()),
            'confirmation_method' => 'manual',
            'confirm' => true,
            'setup_future_usage' => 'off_session',
            'use_stripe_sdk' => true,
            'description' => Utils::getPaymentDescriptionForOrder($Order)
        ]);
    }

    /**
     * Create a Subscription from a (temporary) Order
     *
     * @param AbstractOrder $Order
     * @return string|null - Subscription ID
     *
     * @throws Exception
     */
    public function createSubscription(AbstractOrder $Order): ?string
    {
        return Subscriptions::createSubscription($Order);
    }

    /**
     * Capture subscription amount based on an Invoice
     *
     * @param Invoice $Invoice
     * @return void
     *
     * @throws Exception
     */
    public function captureSubscription(Invoice $Invoice): void
    {
        Subscriptions::billSubscriptionBalance($Invoice);
    }

    /**
     * Cancel a Subscription
     *
     * @param int|string $subscriptionId
     * @param string $reason (optional) - The reason why the subscription is cancelled
     * @return void
     *
     * @throws Exception
     */
    public function cancelSubscription(int|string $subscriptionId, string $reason = ''): void
    {
        Subscriptions::cancelSubscription($subscriptionId, $reason);
    }

    /**
     * Suspend a Subscription
     *
     * This *temporarily* suspends the automated collection of payments until explicitly resumed.
     *
     * @param int|string $subscriptionId
     * @param string|null $note (optional) - Suspension note
     * @return void
     * @throws ApiErrorException
     * @throws StripeException
     */
    public function suspendSubscription(int|string $subscriptionId, string $note = null): void
    {
        Subscriptions::suspendSubscription($subscriptionId);
    }

    /**
     * Resume a suspended Subscription
     *
     * This resumes automated collection of payments of a previously superdense Subscription.
     *
     * @param int|string $subscriptionId
     * @param string|null $note (optional) - Resume note
     * @return void
     * @throws ApiErrorException
     * @throws StripeException
     */
    public function resumeSubscription(int|string $subscriptionId, string $note = null): void
    {
        Subscriptions::resumeSubscription($subscriptionId);
    }

    /**
     * Checks if a subscription is currently suspended
     *
     * @param int|string $subscriptionId
     * @return bool
     */
    public function isSuspended(int|string $subscriptionId): bool
    {
        return Subscriptions::isSuspended($subscriptionId);
    }

    /**
     * Sets a subscription as inactive (on the side of this QUIQQER system only!)
     *
     * IMPORTANT: This does NOT mean that the corresponding subscription at the payment provider
     * side is cancelled. If you want to do this please use cancelSubscription() !
     *
     * @param int|string $subscriptionId
     * @return void
     */
    public function setSubscriptionAsInactive(int|string $subscriptionId): void
    {
        Subscriptions::setSubscriptionAsInactive($subscriptionId);
    }

    /**
     * Can the Subscription of this payment method be edited
     * regarding essential data like invoice frequency, amount etc.?
     *
     * @return bool
     */
    public function isSubscriptionEditable(): bool
    {
        return false;
    }

    /**
     * Check if a Subscription is associated with an order and
     * return its ID (= identification at the payment method side; e.g. PayPal)
     *
     * @param AbstractOrder $Order
     * @return int|string|false - ID or false of no ID associated
     */
    public function getSubscriptionIdByOrder(AbstractOrder $Order): bool|int|string
    {
        try {
            $result = QUI::getDataBase()->fetch([
                'select' => ['stripe_id'],
                'from' => Provider::getStripeBillingSubscriptionsTable(),
                'where' => [
                    'global_process_id' => $Order->getUUID()
                ]
            ]);
        } catch (Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return false;
        }

        if (empty($result)) {
            return false;
        }

        return $result[0]['stripe_id'];
    }

    /**
     * Checks if the subscription is active at the payment provider side
     *
     * @param int|string $subscriptionId
     * @return bool
     */
    public function isSubscriptionActiveAtPaymentProvider(int|string $subscriptionId): bool
    {
        try {
            $data = Subscriptions::getSubscriptionDetails($subscriptionId);
        } catch (Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return true;
        }

        if (empty($data)) {
            QUI\System\Log::addError(
                'Stripe subscription details are empty (ID #' . $subscriptionId . '). Please check manually.'
            );

            return true;
        }

        return $data['status'] === StripeSubscription::STATUS_ACTIVE;
    }

    /**
     * Checks if the subscription is active at QUIQQER
     *
     * @param string|int $subscriptionId - Payment provider subscription ID
     * @return bool
     */
    public function isSubscriptionActiveAtQuiqqer(int|string $subscriptionId): bool
    {
        try {
            $result = QUI::getDataBase()->fetch([
                'select' => ['active'],
                'from' => Provider::getStripeBillingSubscriptionsTable(),
                'where' => [
                    'stripe_id' => $subscriptionId
                ]
            ]);
        } catch (Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return true;
        }

        if (empty($result)) {
            return false;
        }

        return !empty($result[0]['active']);
    }

    /**
     * Get IDs of all subscriptions
     *
     * @param bool $includeInactive (optional) - Include inactive subscriptions [default: false]
     * @return int[]
     */
    public function getSubscriptionIds(bool $includeInactive = false): array
    {
        $where = [];

        if (empty($includeInactive)) {
            $where['active'] = 1;
        }

        try {
            $result = QUI::getDataBase()->fetch([
                'select' => ['stripe_id'],
                'from' => Provider::getStripeBillingSubscriptionsTable(),
                'where' => $where
            ]);
        } catch (Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return [];
        }

        return array_column($result, 'stripe_id');
    }

    /**
     * Get global processing ID of a subscription
     *
     * @param string|int $subscriptionId
     * @return string|false
     */
    public function getSubscriptionGlobalProcessingId(int|string $subscriptionId): bool|string
    {
        try {
            $result = QUI::getDataBase()->fetch([
                'select' => ['global_process_id'],
                'from' => Provider::getStripeBillingSubscriptionsTable(),
                'where' => [
                    'stripe_id' => $subscriptionId
                ]
            ]);
        } catch (Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return false;
        }

        if (empty($result)) {
            return false;
        }

        return $result[0]['global_process_id'];
    }

    /**
     * If the Payment method is a payment gateway, it can return a gateway display
     *
     * @param AbstractOrder $Order
     * @param QUI\ERP\Order\Controls\OrderProcess\Processing $Step
     * @return string
     *
     * @throws QUI\Exception
     * @throws Exception
     */
    public function getGatewayDisplay(AbstractOrder $Order, $Step = null): string
    {
        $type = $this->getPaymentMethodType();

        $Control = new PaymentDisplay([
            'paymentMethod' => $type,
            'infoText' => $this->getPaymentStepInfo()
        ]);

        $Control->setAttribute('Order', $Order);

        $stepTitle = $this->getPaymentStepTitle();
        $Step->setTitle($stepTitle);

        $Engine = QUI::getTemplateManager()->getEngine();

        $Engine->assign([
            'stepTitle' => $stepTitle
        ]);

        $Step->setContent($Engine->fetch(dirname(__FILE__, 3) . '/PaymentDisplay.Header.html'));

        return $Control->create();
    }

    /**
     * Does the payment ONLY support recurring payments (e.g. for subscriptions)?
     *
     * @return bool
     */
    public function supportsRecurringPaymentsOnly(): bool
    {
        return true;
    }
}
